package de.fhhannover.inform.trust.ifmapj;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateException;

import javax.net.ssl.KeyManager;
import javax.net.ssl.KeyManagerFactory;
import javax.net.ssl.TrustManager;
import javax.net.ssl.TrustManagerFactory;

import de.fhhannover.inform.trust.ifmapj.channel.SSRC;
import de.fhhannover.inform.trust.ifmapj.exception.InitializationException;

/**
 * To create a {@link SSRC} one needs to have {@link KeyManager} and
 * {@link TrustManager} instances. This is a helper class to make it
 * easy to get those.
 * 
 * @author aw
 *
 */
public abstract class IfmapJHelper {
	
	/**
	 * Can set the environment variables... Just for convenience.
	 * 
	 * @param tsPath
	 * @param tsPass
	 * @param ksPath
	 * @param ksPass
	 */
	public static void setSystemProperties(String tsPath, String tsPass, String ksPath, String ksPass) {
		if (tsPath != null)
			System.setProperty("javax.net.ssl.trustStore", tsPath);
		
		if (tsPass != null)
			System.setProperty("javax.net.ssl.trustStorePassword",tsPass);
		
		if (ksPath != null)
			System.setProperty("javax.net.ssl.keyStore", ksPath);
		
		if (ksPass != null)
			System.setProperty("javax.net.ssl.keyStorePassword",ksPass);
	}
	
	/**
	 * @return an array of KeyManager instances, based on the javax.net.ssl.keyStore
	 * 			and javax.net.ssl.keyStorePassword environment variables.
	 * @throws InitializationException
	 */
	public static KeyManager[] getKeyManagers() throws InitializationException {
		String path = System.getProperty("javax.net.ssl.keyStore");
		String pass = System.getProperty("javax.net.ssl.keyStorePassword");
		
		if (path == null || pass == null)
			throw new InitializationException("javax.net.ssl.keyStore / " +
					"javax.net.ssl.keyStorePassword not set");
		
		return getKeyManagers(path, pass);
	}

	/**
	 * @param keyStore path to trustStore
	 * @param pass password for keyStore
	 * @return an array of KeyManager instances, based on the given parameters
	 * @throws InitializationException
	 */
	public static KeyManager[] getKeyManagers(String keyStore, String pass) throws InitializationException {
		if (keyStore == null || pass == null)
			throw new NullPointerException("parameters null?");
		try {
			KeyManagerFactory kmf = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm());
			KeyStore ks = loadKeyStore(keyStore, pass);
			kmf.init(ks, pass.toCharArray());
			return kmf.getKeyManagers();
		} catch (Exception e) {
			throw new InitializationException(e.getMessage());
		}
	}

	/**
	 * @return an array of TrustManager instances, based on the javax.net.ssl.trustStore
	 * 			and javax.net.ssl.trustStorePassword environment variables.
	 * @throws InitializationException
	 */
	public static TrustManager[] getTrustManagers() throws InitializationException {
		String path = System.getProperty("javax.net.ssl.trustStore");
		String pass = System.getProperty("javax.net.ssl.trustStorePassword");
		
		if (path == null || pass == null)
			throw new InitializationException("javax.net.ssl.trustStore / " +
					"javax.net.ssl.trustStorePassword not set");
		
		return getTrustManagers(path, pass);
			
	}
	
	/**
	 * @param trustStore path to trustStore
	 * @param pass password for trustStore
	 * @return an array of TrustManager instances, based on the given parameters
	 * @throws InitializationException
	 */
	public static TrustManager[] getTrustManagers(String trustStore, String pass) throws InitializationException {
		if (trustStore == null || pass == null)
			throw new NullPointerException("parameters null?");
		try {
			TrustManagerFactory tmf = TrustManagerFactory.getInstance(TrustManagerFactory.getDefaultAlgorithm());
			KeyStore ks = loadKeyStore(trustStore, pass);
			tmf.init(ks);
			return tmf.getTrustManagers();
		} catch (Exception e) {
			throw new InitializationException(e.getMessage());
		}
		
	}
	
	/**
	 * Helper to load a {@link KeyStore} instance.
	 * 
	 * @param path
	 * @param pass
	 * @return
	 * @throws NoSuchAlgorithmException
	 * @throws CertificateException
	 * @throws IOException
	 * @throws KeyStoreException
	 */
	private static KeyStore loadKeyStore(String path, String pass) 
			throws NoSuchAlgorithmException, CertificateException, IOException, KeyStoreException {
		KeyStore ks = KeyStore.getInstance(KeyStore.getDefaultType());
		InputStream is = new FileInputStream(new File(path));
		ks.load(is, pass.toCharArray());
		return ks;
	}
}
