package de.fhhannover.inform.trust.ifmapj.channel;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.security.SecureRandom;
import java.util.zip.GZIPOutputStream;

import javax.net.ssl.KeyManager;
import javax.net.ssl.SSLContext;
import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.TrustManager;

import de.fhhannover.inform.trust.ifmapj.binding.BindingFactory;
import de.fhhannover.inform.trust.ifmapj.binding.Marshaller;
import de.fhhannover.inform.trust.ifmapj.binding.Unmarshaller;
import de.fhhannover.inform.trust.ifmapj.exception.CommunicationException;
import de.fhhannover.inform.trust.ifmapj.exception.InitializationException;
import de.fhhannover.inform.trust.ifmapj.log.IfmapJLog;
import de.fhhannover.inform.trust.ifmapj.messages.RequestFactoryImpl;
import de.fhhannover.inform.trust.ifmapj.messages.RequestFactoryInternal;

/**
 * Encapsulate functionality needed by both {@link SSRC} and {@link ARC}
 * implementations.
 * 
 * @author aw
 *
 */
abstract class Channel implements IfmapChannel {
	
	/**
	 * represents the handler used for communication to be used
	 */
	private final CommunicationHandler mCommunicationHandler;

	/**
	 * Represents the url in string form
	 */
	private final String mUrlStr;
	
	/**
	 * user to be used for basic authentication, null for certificate-based
	 */
	private final String mUser;
	
	/**
	 * password to be used for basic authentication, null for certificate-based
	 */
	private final String mPassword;
	
	/**
	 * whether this channel uses basic authentication or certificate-based
	 * authentication
	 */
	private final boolean mBasicAuth;
	
	/**
	 * request factory used by {@link SSRC} and {@link ARC} implementation.
	 */
	protected final RequestFactoryInternal mRequestFactory;
	
	/**
	 * {@link Marshaller} implementation to be used by subclasses.
	 */
	protected final Marshaller mMarshaller;
	
	/**
	 * {@link Unmarshaller} implementation to be used by subclasses.
	 */
	protected final Unmarshaller mUnmarshaller;

	/**
	 * Construct a {@link Channel} object in the most general form.
	 * 
	 * @param url
	 * @param user
	 * @param pass
	 * @param ks the keystore managers to be used
	 * @param ts the truststore managers to be used
	 * @throws InitializationException
	 */
	private Channel(String url, String user, String pass, KeyManager[] kms, TrustManager[] tms)
			throws InitializationException  {
		
		if (url == null)
			throw new InitializationException("URL not allowed to be null");
	
		// check only trustmanager
		if (tms == null)
			throw new InitializationException("keystore and truststore need " +
					"to be set");
	
		// If basic authentication  is used, make sure user and pass are set 
		if ((user != null && pass == null) || (user == null && pass != null))
			throw new InitializationException("One basic auth parameter is null");
		
		
		mUrlStr = url;
		mUser = user;
		mPassword = pass;
		mBasicAuth = !(mUser == null && mPassword == null);
		
		mRequestFactory = new RequestFactoryImpl();
		mMarshaller = BindingFactory.newMarshaller();
		mUnmarshaller = BindingFactory.newUnmarshaller();
		SSLSocketFactory sslSocketFactory = initSslSocketFactory(kms, tms);
		mCommunicationHandler = CommunicationHandlerFactory.newHandler(
				url, user, pass, sslSocketFactory);
	}


	/**
	 * Used for basic-authentication
	 * 
	 * @param url
	 * @param user
	 * @param pass
	 * @param tms
	 * @throws InitializationException
	 */
	Channel(String url, String user, String pass, TrustManager[] tms) throws InitializationException {
		this(url, user, pass, null, tms);
	}
	
	/**
	 * Used for cert-based authentication
	 * @param url
	 * @param kms
	 * @param tms
	 * @throws InitializationException
	 */
	Channel(String url, KeyManager[] kms, TrustManager[] tms)
			throws InitializationException {
		this(url, null, null, kms, tms);
	}

	protected InputStream doHttpRequest(InputStream is) throws CommunicationException {
		
		// if we use gzip, transform is into a gzipped form.
		if (usesGzip())
			is = makeGzippedInputStream(is);

		return mCommunicationHandler.doRequest(is);
	}
	
	public String getUrl() {
		return mUrlStr;
	}
	
	public String getUser() {
		return mUser;
	}
	
	public String getPassword() {
		return mPassword;
	}
	
	@Override
	public void setGzip(boolean useGzip) {
		mCommunicationHandler.setGzip(useGzip);
	}

	@Override
	public boolean usesGzip() {
		return mCommunicationHandler.usesGzip();
	}
	
	/**
	 * @return whether the channel uses basic authentication
	 */
	public boolean isBasicAuth() {
		return mBasicAuth;
	}

	@Override
	public void closeTcpConnection() throws CommunicationException {
		mCommunicationHandler.closeTcpConnection();
	}
	
	
	/**
	 * I guess we could make this all easier by just using the default
	 * keystore... 
	 * 
	 * @throws InitializationException
	 */
	private SSLSocketFactory initSslSocketFactory(KeyManager[] kms, TrustManager[] tms)
			throws InitializationException {
		
		SSLContext ctx = null;
		
		if (!isBasicAuth() && kms == null)
			throw new InitializationException("certificate-based auth needs a KeyManager");
		
		try {
			ctx = SSLContext.getInstance("TLS");
			ctx.init(kms, tms, new SecureRandom());
		} catch (Exception e) {
			/* catch all */
			IfmapJLog.error("Could not initialize SSLSocketFactory ["
			                                      + e.getMessage() + "]");
			throw new InitializationException(e);
		}
		return ctx.getSocketFactory();
	}

	/**
	 * Use gzip to compress the input stream, we do copying here, but
	 * I'm not sure whether there is a way around it anyway...
	 * 
	 * @param is
	 * @return
	 * @throws CommunicationException if something goes wrong while working on
	 *         the {@link InputStream} instances.
	 */
	private InputStream makeGzippedInputStream(InputStream is) throws CommunicationException {
		try {
		int next = -1;
			ByteArrayOutputStream baos = new ByteArrayOutputStream();
			OutputStream os = new GZIPOutputStream(baos);
			
			while ((next = is.read()) >= 0)
				os.write(next);
			
			baos.close();
			os.close();
			is.close();
			return new ByteArrayInputStream(baos.toByteArray());
		
		} catch (IOException e) {
			throw new CommunicationException(e);
		}
	}


}
