package de.fhhannover.inform.trust.ifmapj.channel;

import java.io.ByteArrayInputStream;
import java.io.DataInput;
import java.io.DataInputStream;
import java.io.DataOutput;
import java.io.DataOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;
import java.util.Map;
import java.util.zip.GZIPInputStream;

import javax.net.ssl.SSLSocket;
import javax.net.ssl.SSLSocketFactory;

import util.Base64;
import de.fhhannover.inform.trust.ifmapj.exception.CommunicationException;
import de.fhhannover.inform.trust.ifmapj.exception.InitializationException;
import de.fhhannover.inform.trust.ifmapj.log.IfmapJLog;

/**
 * Attempt to create a {@link CommunicationHandler} implementation to be
 * independent of any extra library... Meaning making usage only of standard
 * JAVA classes.
 * 
 * @author aw
 */
class JavaCommunicationHandler implements CommunicationHandler {
	
	private final SSLSocketFactory mSocketFactory;
	private final String mAuthHeader;
	private final boolean mBasicAuth;
	private final String mPath;
	private final URL mUrl;
	private final int mPort;
	
	private SSLSocket mSocket;
	private OutputStream mOutputStream;
	private InputStream mInputStream;
	private DataOutput mDataOutputStream;
	private DataInput mDataInputStream;
	private boolean mGzip;
	
	private byte[] mBuffer;
	
	private class StatusLine {
		public int statusCode;
		public String reason;
	}
	
	JavaCommunicationHandler(String url, String user, String pass,
			SSLSocketFactory sslSocketFactory) throws InitializationException {
		
		if (url == null)
			throw new NullPointerException("url is null");
		
		if (sslSocketFactory == null)
			throw new NullPointerException("sslSocketFactory is null");
		
		try {
			mUrl = new URL(url);
		} catch (MalformedURLException e) {
			throw new InitializationException("Malformed URL [" 
					+ e.getMessage() + "]");
		}

		// fix up the path in case non is given
		mPath = (mUrl.getPath().length() == 0) ? "/" : mUrl.getPath();
	
		mPort = (mUrl.getPort() == -1) ? mUrl.getDefaultPort() : mUrl.getPort();
	
		mBasicAuth = !(user == null || pass == null);
		mAuthHeader = "Basic " + Base64.encodeToString((user + ":" + pass).getBytes(), false);
		
		mSocketFactory = sslSocketFactory;
		
		mBuffer = new byte[1024];
	}
	
	

	@Override
	public InputStream doRequest(InputStream is) throws CommunicationException {
		try {
			boolean gzipCompressed = false;
			StatusLine status = null;
			Map<String, String> headers = null;
			InputStream isret = null;
			int length = is.available();
			
			prepareCommunication();
			
			writePostLine();
			writeContentTypeHeaders();
			
			if (mBasicAuth)
				writeAuthHeader();
			
			if (usesGzip())
				writeGzipHeaders();
			
			writeHeaderLength(length);
			writeHeaderEnding();
			
			sendHttpBody(is);

			status = readStatusLine();
			checkStatus(status);
			
			headers = readHeaders();
			length = getContentLength(headers);
			gzipCompressed = isGzipCompressed(headers);
			
			isret = readHttpBody(length);
			
			if (gzipCompressed)
				isret = new GZIPInputStream(isret);
			
			return isret;
		} catch (IOException e) {
			throw new CommunicationException(e);
		}
	}

	@Override
	public void setGzip(boolean gzip) {
		mGzip = gzip;
	}

	@Override
	public boolean usesGzip() {
		return mGzip;
	}

	/*
	 * Try closing it under all circumstances. Throw an Exception of the first
	 * error encountered.
	 * 
	 * It's soooo ugly.
	 * 
	 * (non-Javadoc)
	 * @see de.fhhannover.inform.trust.ifmapj.channel.CommunicationHandler#closeTcpConnection()
	 */
	@Override
	public void closeTcpConnection() throws CommunicationException {
		IOException tmpException = null;

		try {
			if (mInputStream != null) mInputStream.close();
		} catch (IOException e) { 
			if (tmpException == null) tmpException = e;
		} finally {
			mInputStream = null;
		}
		
		try {
			if (mOutputStream != null) mOutputStream.close();
		} catch (IOException e) { 
			if (tmpException != null) tmpException = e;
		} finally {
			mOutputStream = null;
		}
		
		try {
			if (mSocket != null) mSocket.close();
		} catch (IOException e) { 
			if (tmpException == null) tmpException = e;
		} finally {
			mSocket = null;
		}
		
		if (tmpException != null)
			throw new CommunicationException(tmpException);

		mDataOutputStream = null;
		mDataInputStream = null;
	}

	/**
	 * Initialize socket and streams if not yet done.
	 * 
	 * @throws IOException
	 */
	private void prepareCommunication() throws IOException {
		if (mSocket == null) {
			mSocket = (SSLSocket)mSocketFactory.createSocket(mUrl.getHost(), mPort);
			mSocket.setTcpNoDelay(true);
			mSocket.setNeedClientAuth(true);
			mOutputStream = mSocket.getOutputStream();
			mInputStream = mSocket.getInputStream();
			mDataOutputStream =  new DataOutputStream(mOutputStream);
			mDataInputStream = new DataInputStream(mInputStream);
		}
	}

	private boolean isGzipCompressed(Map<String, String> headers) {
		String encoding = null;
		encoding = headers.get("Content-Encoding");
		if (encoding == null)
			encoding = headers.get("Content-encoding");

		if (encoding == null)
			encoding = headers.get("content-encoding");
		
		if (encoding != null)
			return encoding.contains("gzip");
		
		return false;
	}

	private int getContentLength(Map<String, String> headers) 
			throws CommunicationException {
		int ret;
		String lengthHdr = null;
		
		lengthHdr = headers.get("Content-Length");
		
		if (lengthHdr == null)
			lengthHdr = headers.get("Content-length");

		if (lengthHdr == null)
			lengthHdr = headers.get("content-length");
			
		if (lengthHdr == null)
			throw new CommunicationException("No Content-Length header found");
	
		
		try {
			ret = Integer.parseInt(lengthHdr);
		} catch (NumberFormatException e) {
			throw new CommunicationException("Content-Length invalid: \"" + lengthHdr 
					+ "\"");
		}
		return ret;
	}

	private Map<String, String> readHeaders() throws CommunicationException, IOException  {
		
		String line = mDataInputStream.readLine();
		HashMap<String, String> ret = new HashMap<String, String>();
		
		while (line != null && line.length() > 0) {
			String fields[] = line.split(":", 2);
			if (fields.length < 2)
				throw new CommunicationException("Invalid Header Received: "  + line);

			// trim a bit
			fields[0] = fields[0].replaceAll("^\\s+", "");
			fields[0] = fields[0].replaceAll("\\s+$", "");
			fields[1] = fields[1].replaceAll("^\\s+", "");
			fields[1] = fields[1].replaceAll("\\s+$", "");
			
			ret.put(fields[0], fields[1]);
			
			line = mDataInputStream.readLine();
		}
		
		if (line == null)
			throw new CommunicationException("Unexpected EOF reached");
		
		return ret;
	}

	private StatusLine readStatusLine() throws CommunicationException, IOException {
		StatusLine ret = new StatusLine();
		String line = mDataInputStream.readLine();
		if (line == null)
			throw new CommunicationException("No status line received");
		
		String fields[] = line.split(" ", 3);
		if (fields.length < 2)
			throw new CommunicationException("Bad status line received");
		String proto = fields[0];
		
		if (!proto.equals("HTTP/1.1"))
			throw new  CommunicationException("Communication not HTTP/1.1");

		try {
			ret.statusCode = Integer.parseInt(fields[1]);
		} catch (NumberFormatException e) {
			throw new CommunicationException("Bad status code received");
		}
	
		if (fields.length == 3)
			ret.reason = fields[2];
		
		return ret;
	}
	
	private void writeHeaderLength(int length) throws IOException {
		writeHeaderLine("Content-Length", "" + length);
	}

	private void writeGzipHeaders() throws IOException {
		writeHeaderLine("Accept-Encoding", "gzip");
		writeHeaderLine("Content-Encoding", "gzip");
	}

	private void writeAuthHeader() throws IOException {
		writeHeaderLine("Authorization", mAuthHeader);
	}

	private void writeContentTypeHeaders() throws IOException {
		writeHeaderLine("Accept-Type", "application/soap+xml");
		writeHeaderLine("Content-Type", "application/soap+xml");
	}

	private void writePostLine() throws IOException {
		writeLine("POST " + mPath + " HTTP/1.1");
	}
	
	private void writeHeaderLine(String key, String value) throws IOException {
		writeLine(key + ": " + value);
	}

	private void writeHeaderEnding() throws IOException {
		writeLine("");
		mOutputStream.flush();
	}
	
	private void writeLine(String line) throws IOException {
		mDataOutputStream.write((line + "\r\n").getBytes());
	}

	private InputStream readHttpBody(int length) throws IOException {
		allocateBuffer(length);
		readStreamIntoBuffer(mInputStream, length);
		return new ByteArrayInputStream(mBuffer, 0, length);
	}

	private void checkStatus(StatusLine status) throws CommunicationException {
		if (status.statusCode != 200) {
			IfmapJLog.warn("HTTP Status Code: " 
					+ status.statusCode + " "
					+ status.reason);
			throw new CommunicationException("HTTP Status Code: "
					+ status.statusCode + " " 
					+ status.reason);
		}
	}

	private void sendHttpBody(InputStream is) throws IOException {
		int length = is.available();
		allocateBuffer(length);
		readStreamIntoBuffer(is, length);
		sendBufferContents(length);
	}

	private void sendBufferContents(int length) throws IOException {
		mOutputStream.write(mBuffer, 0, length);
		mOutputStream.flush();
	}

	/**
	 * Helper to read length bytes into {@link JavaCommunicationHandler#mBuffer}.
	 * 
	 * @param is
	 * @param length
	 * @throws IOException
	 */
	private void readStreamIntoBuffer(InputStream is, int length) throws IOException {
		int read = 0;
		int ret;
		
		while (read < length) {
			ret = is.read(mBuffer, read, length - read);
			if (ret == -1)
				throw new IOException("Stream exception");
			read += ret;
		}
	}

	/**
	 * Check if our current buffer is too small, and if yes, allocate some
	 * more
	 * 
	 * @param length
	 */
	private void allocateBuffer(int length) {
		if (length > mBuffer.length)
			mBuffer = new byte[length];
	}

}
